@@ -4,7 +4,7 @@ Juju Client
 
 Seriously Alpha. Works now, but API *will* change.
 
-A simple synchronous python client for the juju-core/gojuju websocket api.
+A simple synchronous python client for the juju-core websocket api.
 
 Example Usage::
 
@@ -42,16 +42,23 @@ Upstream/Server
   - need proper status output, or other introspection beyond AllWatcher
   - deploy local charm
   - bad constraints fail silently
+
   - need terminate machine api
   - clarify usage/working of env annotation
 """
 # License: GPLv3
 # Author: Kapil Thangavelu <kapil.foss@gmail.com>
 
+from base64 import b64encode
 from contextlib import contextmanager
+
+import errno
+import httplib
 import json
 import pprint
+import time
 import signal
+import socket
 import StringIO
 import logging
 import websocket
@@ -99,6 +106,9 @@ class EnvError(Exception):
     def __init__(self, error):
         self.error = error
         self.message = error['Error']
+        # Call the base class initializer so that this exception can be pickled
+        # (see http://bugs.python.org/issue1692335).
+        super(EnvError, self).__init__(error)
 
     def __str__(self):
         stream = StringIO.StringIO()
@@ -107,11 +117,19 @@ class EnvError(Exception):
             stream.getvalue())
 
 
+class Jobs(object):
+    HostUnits = "JobHostUnits"
+    ManageEnviron = "JobManageEnviron"
+    ManageState = "JobManageState"
+
+
 class RPC(object):
 
     _auth = False
     _request_id = 0
     _debug = False
+    _reconnect_params = None
+    conn = None
 
     def _rpc(self, op):
         if not self._auth and not op.get("Request") == "Login":
@@ -129,22 +147,59 @@ class RPC(object):
             log.debug("rpc response:\n%s" % (json.dumps(result, indent=2)))
 
         if 'Error' in result:
-            # print "raw", op['Request'], raw
             # The backend disconnects us on err, bug: http://pad.lv/1160971
             self.conn.connected = False
             raise EnvError(result)
         return result['Response']
 
+    def login(self, password, user="user-admin", **ignore):
+        """Login gets shared to watchers for reconnect."""
+        if self.conn and self.conn.connected and self._auth:
+            raise AlreadyConnected()
+        # Store for constructing separate authenticated watch connections.
+        self._creds = {'password': password, 'user': user}
+        result = self._rpc(
+            {"Type": "Admin", "Request": "Login",
+             "Params": {"AuthTag": user, "Password": password}})
+        self._auth = True
+        return result
+
+    def set_reconnect_params(self, params):
+        self._reconnect_params = params
+
+    def reconnect(self):
+        if self.conn:
+            self._auth = False
+            self.conn.close()
+        if not self._reconnect_params:
+            return False
+
+        log.info("Reconnecting client")
+        while True:
+            try:
+                self.conn = websocket.create_connection(
+                    self._reconnect_params['url'],
+                    origin=self._reconnect_params['origin'])
+                break
+            except socket.error as err:
+                if not err.errno in (
+                        errno.ETIMEDOUT, errno.ECONNREFUSED, errno.ECONNRESET):
+                    raise
+                time.sleep(1)
+                continue
+        self.login(**self._reconnect_params)
+        return True
+
 
 class Watcher(RPC):
 
     _auth = True
 
-    def __init__(self, conn):
+    def __init__(self, conn, auto_reconnect=True):
         self.conn = conn
         self.watcher_id = None
         self.running = False
-
+        self.auto_reconnect = auto_reconnect
         # For debugging, attach the wrapper
         self.context = None
 
@@ -162,20 +217,38 @@ class Watcher(RPC):
             self.start()
         if not self.running:
             raise StopIteration("Stopped")
-        result = self._rpc({
-            'Type': 'AllWatcher',
-            'Request': 'Next',
-            'Id': self.watcher_id})
+        try:
+            result = self._rpc({
+                'Type': 'AllWatcher',
+                'Request': 'Next',
+                'Id': self.watcher_id})
+        except EnvError, e:
+            if "state watcher was stopped" in e.message:
+                if not self.auto_reconnect:
+                    raise
+                if not self.reconnect():
+                    raise
+                return self.next()
+            raise
         return result['Deltas']
 
+    def reconnect(self):
+        self.watcher_id = None
+        self.running = False
+        return super(Watcher, self).reconnect()
+
     def stop(self):
         if not self.conn.connected:
             return
-        result = self._rpc({
-            'Type': 'AllWatcher',
-            'Request': 'Stop',
-            'Id': self.watcher_id})
+        try:
+            result = self._rpc({
+                'Type': 'AllWatcher', 'Request': 'Stop',
+                'Id': self.watcher_id})
+        except (EnvError, socket.error):
+            # We're about to close the connection.
+            result = None
         self.conn.close()
+        self.watcher_id = None
         self.running = False
         return result
 
@@ -246,24 +319,176 @@ class Environment(RPC):
         if self.conn.connected:
             self.conn.close()
 
-    def login(self, password, user="user-admin"):
-        if self.conn and self.conn.connected and self._auth:
-            raise AlreadyConnected()
-        # Store for constructing separate authenticated watch connections.
-        self._creds = {'password': password, 'user': user}
-        self._rpc({"Type": "Admin", "Request": "Login",
-                   "Params": {"AuthTag": user, "Password": password}})
-        self._auth = True
+    # Charm ops
+    def add_local_charm(self, charm_file, series, size=None):
+        """Add a local charm to an environment.
 
+        Uses an https endpoint at the same host:port as the wss.
+        Supports large file uploads.
+        """
+        endpoint = self.endpoint.replace('wss://', '')
+        host, port = endpoint.split(':')
+        conn = httplib.HTTPSConnection(host, port)
+        path = "/charms?series=%s" % (series)
+        headers = {
+            'Content-Type': 'application/zip',
+            'Authorization': 'Basic %s' % b64encode(
+                '%(user)s:%(password)s' % (self._creds))}
+        # Specify if its a psuedo-file object,
+        # httplib will try to stat non strings.
+        if size:
+            headers['Content-Length'] = size
+        conn.request("POST", path, charm_file, headers)
+        response = conn.getresponse()
+        result = json.loads(response.read())
+        if not response.status == 200:
+            raise EnvError(result)
+        return result
+
+    def add_charm(self, charm_url):
+        return self._rpc(
+            {"Type": "Client",
+             "Request": "AddCharm",
+             "Params": {"URL": charm_url}})
+
+    # Environment operations
     def info(self):
         return self._rpc({
             "Type": "Client",
             "Request": "EnvironmentInfo"})
 
     def status(self):
-        # Status is currently broken, only reports machine ids.
+        # Status via api is currently broken, only reports machine ids,
+        # use an all watch with status translator to get something usable.
         return self._rpc({"Type": "Client", "Request": "Status"})
 
+    def get_charm(self, charm_url):
+        return self._rpc(
+            {"Type": "Client",
+             "Request": "CharmInfo",
+             "Params": {
+                 "CharmURL": charm_url}})
+
+    # Environment
+    def get_env_constraints(self):
+        return self._rpc({
+            "Type": "Client",
+            "Request": "GetEnvironmentConstraints"})
+
+    def set_env_constraints(self, constraints):
+        return self._rpc({
+            "Type": "Client",
+            "Request": "SetEnvironmentConstraints",
+            "Params": {}})
+
+    def get_env_config(self):
+        return self._rpc({
+            "Type": "Client",
+            "Request": "EnvironmentGet"})
+
+    def set_env_config(self, config):
+        return self._rpc({
+            "Type": "Client",
+            "Request": "EnvironmentSet",
+            "Params": {"Config": config}})
+
+    # Machine ops
+    def add_machine(self, series="", constraints=None,
+                    machine_spec="", parent_id="", container_type=""):
+
+        """Allocate a new machine from the iaas provider.
+        """
+        if machine_spec:
+            err_msg = "Cant specify machine spec with container_type/parent_id"
+            assert not (parent_id or container_type), err_msg
+            parent_id, container_type = machine_spec.split(":", 1)
+
+        params = dict(
+            Series=series,
+            Constraints=self._prepare_constraints(constraints),
+            ContainerType=container_type,
+            ParentId=parent_id,
+            Jobs=[Jobs.HostUnits])
+        return self.add_machines([params])['Machines'][0]
+
+    def add_machines(self, machines):
+        """Allocate multiple machines from the iaas provider.
+
+        See add_machine for format of parameters.
+        """
+        return self._rpc({
+            "Type": "Client",
+            "Request": "AddMachines",
+            "Params": {
+                "MachineParams": machines}})
+
+    def register_machine(self, instance_id, nonce, series, hardware, addrs):
+        """Register/Enlist a machine into an environment state.
+
+        The machine will need to have tools installed and subsequently
+        connect to the state server with the given nonce
+        credentials. The machine_config method can be used to
+        construct a suitable set of commands.
+
+        Parameters:
+
+        nonce: is the initial password for the new machine.
+        addrs: list of ip addresses for the machine.
+        hw: is the hardware characterstics of the machine, applicable keys.
+         - Arch
+         - Mem
+         - RootDisk size
+         - CpuCores
+         - CpuPower
+         - Tags
+        """
+        params = dict(
+            Series=series,
+            InstanceId=instance_id,
+            Jobs=[Jobs.HostUnits],
+            HardwareCharacteristics=hardware,
+            Addrs=addrs,
+            Nonce=nonce)
+        return self.register_machines([params])['Machines'][0]
+
+    def register_machines(self, machines):
+        return self._rpc({
+            "Type": "Client",
+            "Request": "InjectMachines",
+            "Params": {
+                "MachineParams": machines}})
+
+    def destroy_machines(self, machine_ids, force=False):
+        params = {"MachineNames": machine_ids}
+        if force:
+            params["Force"] = True
+        return self._rpc({
+            "Type": "Client",
+            "Request": "DestroyMachines",
+            "Params": params})
+
+    def provisioning_script(self, machine_id, nonce,
+                            data_dir="/var/lib/juju", disable_apt=False):
+        return self._rpc({
+            "Type": "Client",
+            "Request": "ProvisioningScript",
+            "Params": {
+                "MachineId": machine_id,
+                "Nonce": nonce,
+                "DataDir": data_dir,
+                "DisablePackageCommands": disable_apt}})
+
+    def machine_config(self, machine_id, series, arch):
+        """Return information needed to render cloudinit for a machine.
+        """
+        return self._rpc({
+            "Type": "Client",
+            "Request": "MachineConfig",
+            "Params": {
+                "MachineId": machine_id,
+                "Series": series,
+                "Arch": arch}})
+
     # Watch Wrapper methods
     def get_stat(self):
         """A status emulator using the watch api, returns immediately.
@@ -286,31 +511,35 @@ class Environment(RPC):
         watch = self.get_watch(timeout)
         return WaitForNoMachines(watch).run(callback)
 
-    def get_watch(self, timeout=None):
+    def get_watch(self, timeout=None, connection=None, watch_class=None):
         # Separate conn per watcher to keep sync usage simple, else we have to
         # buffer watch results with requestid dispatch. At the moment
         # with the all watcher, an app only needs one watch, which is likely to
         # change to discrete watches on individual bits.
-        watch_env = Environment(self.endpoint)
-        watch_env.login(**self._creds)
+        if connection is None:
+            watch_env = Environment(self.endpoint)
+            watch_env.login(**self._creds)
+        else:
+            watch_env = connection
+
+        p = dict(self._creds)
+        p.update({'url': self.endpoint, 'origin': self.endpoint})
         if timeout is not None:
-            watcher = TimeoutWatcher(watch_env.conn)
+            if watch_class is None:
+                watch_class = TimeoutWatcher
+            watcher = watch_class(watch_env.conn)
             watcher.set_timeout(timeout)
         else:
-            watcher = Watcher(watch_env.conn)
+            if watch_class is None:
+                watch_class = Watcher
+            watcher = watch_class(watch_env.conn)
+        watcher.set_reconnect_params(p)
         self._watches.append(watcher)
         watcher.start()
         return watcher
 
     watch = get_watch
 
-    def get_charm(self, charm_url):
-        return self._rpc(
-            {"Type": "Client",
-             "Request": "CharmInfo",
-             "Params": {
-                 "CharmURL": charm_url}})
-
     def _prepare_strparams(self, d):
         r = {}
         for k, v in d.items():
@@ -342,7 +571,7 @@ class Environment(RPC):
 
     # Service
     def deploy(self, service_name, charm_url, num_units=1,
-               config=None, constraints=None):
+               config=None, constraints=None, machine_spec=None):
         """Deploy a charm
 
         Does not support local charms
@@ -363,7 +592,8 @@ class Environment(RPC):
                  "CharmURL": charm_url,
                  "NumUnits": num_units,
                  "Config": svc_config,
-                 "Constraints": svc_constraints}})
+                 "Constraints": svc_constraints,
+                 "ToMachineSpec": machine_spec}})
 
     def set_config(self, service_name, config):
         assert isinstance(config, dict)
@@ -375,6 +605,27 @@ class Environment(RPC):
                 "ServiceName": service_name,
                 "Options": svc_config}})
 
+    def unset_config(self, service_name, config_keys):
+        """Unset configuration values of a service to restore charm defaults.
+        """
+        return self._rpc({
+            "Type": "Client",
+            "Request": "ServiceUnset",
+            "Params": {
+                "ServiceName": service_name,
+                "Options": config_keys}})
+
+    def set_charm(self, service_name, charm_url, force=False):
+        """Set the charm url for a service.
+        """
+        return self._rpc({
+            "Type": "Client",
+            "Request": "ServiceSetCharm",
+            "Params": {
+                "ServiceName": service_name,
+                "CharmUrl": charm_url,
+                "Force": force}})
+
     def get_service(self, service_name):
         """Returns dict of Charm, Config, Constraints, Service keys.
 
@@ -390,12 +641,7 @@ class Environment(RPC):
                  "ServiceName": service_name}})
 
     def get_config(self, service_name):
-        """Returns dict of Charm, Config, Constraints, Service keys.
-
-        Charm -> charm used by service
-        Service -> service name
-        Config -> Currently configured options and descriptions
-        Constraints -> Constraints set on service (not environment inherited).
+        """Returns service configuration.
         """
         return self.get_service(service_name)['Config']
 
@@ -414,6 +660,27 @@ class Environment(RPC):
                  "ServiceName": service_name,
                  "Constraints": self._prepare_constraints(constraints)}})
 
+    def update_service(self, service_name, charm_url="", force_charm_url=False,
+                       min_units=None, settings=None, constraints=None):
+        """Update a service.
+
+        Can update a service's charm, modify configuration, constraints,
+        and the number of units.
+        """
+        svc_config = {}
+        if settings:
+            svc_config = self._prepare_strparams(settings)
+
+        return self._rpc(
+            {"Type": "Client",
+             "Request": "SetServiceConstraints",
+             "Params": {
+                 "ServiceName": service_name,
+                 "CharmUrl": charm_url,
+                 "MinUnits": min_units,
+                 "SettingsStrings": svc_config,
+                 "Constraints": self._prepare_constraints(constraints)}})
+
     def destroy_service(self, service_name):
         return self._rpc({
             "Type": "Client",
@@ -435,6 +702,17 @@ class Environment(RPC):
             "Params": {
                 "ServiceName": service_name}})
 
+    def valid_relation_names(self, service_name):
+        """All possible relation names of a service.
+
+        Per its charm metadata.
+        """
+        return self._rpc({
+            "Type": "Client",
+            "Request": "ServiceCharmRelations",
+            "Params": {
+                "ServiceName": service_name}})
+
     # Units
     def add_units(self, service_name, num_units=1):
         return self._rpc({
@@ -444,6 +722,17 @@ class Environment(RPC):
                 "ServiceName": service_name,
                 "NumUnits": num_units}})
 
+    def add_unit(self, service_name, machine_spec=None):
+        params = {
+            "ServiceName": service_name,
+            "NumUnits": 1}
+        if machine_spec:
+            params["ToMachineSpec"] = machine_spec
+        return self._rpc({
+            "Type": "Client",
+            "Request": "AddServiceUnits",
+            "Params": params})
+
     def remove_units(self, unit_names):
         return self._rpc({
             "Type": "Client",
@@ -459,6 +748,16 @@ class Environment(RPC):
                 "UnitName": unit_name,
                 "Retry": retry}})
 
+    # Multi-context
+
+    def get_public_address(self, target):
+        # Return the public address of the machine or unit.
+        return self._rpc({
+            "Type": "Client",
+            "Request": "PublicAddress",
+            "Params": {
+                "Target": target}})
+
     # Annotations
     def set_annotation(self, entity, entity_type, annotation):
         """
@@ -574,9 +873,11 @@ class StatusTranslator(object):
         'Status': 'agent-state',
         "MachineId": "Machine",
         'CharmURL': 'charm',
-        "Number": 'port'
+        'StatusInfo': 'agent-state-info',
+        "Number": 'port',
+        "Ports": "open-ports"
     }
-    remove_keys = set(['Life', "PrivateAddress"])
+    remove_keys = set(['Life', "PrivateAddress", "MinUnits"])
     skip_empty_keys = set(['StatusInfo', "Ports"])
 
     def run(self, watch):
@@ -623,8 +924,7 @@ class StatusTranslator(object):
         ports = d.pop('Ports')
         tports = d.setdefault('Ports', [])
         for p in ports:
-            tports.append(self._translate(p))
-
+            tports.append("%s/%s" % (p['Number'], p['Protocol']))
         svc_units[name] = self._translate(d)
 
     def _service(self, d):
@@ -671,7 +971,6 @@ def main():
     env.deploy("test-blog", "cs:wordpress")
     env.deploy("test-db", "cs:mysql")
     env.add_relation("test-db", "test-blog")
-    config = env.get_config("test-blog")
 
     print "waiting for changes for 30s"
     watcher.set_timeout(30)
